# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..utils import dict_add, url_merge_r, url_merge
from .common import ItemExtractor, ForumExtractor
from .identify import identify_vbulletin
from ..exceptions import AlreadyVisitedError


class vbulletin(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    r"/([0-9]+)-[^/]*/?$",
                    1,
                ),
                (
                    r"/.*([?&](t=)?|-)([0-9]+)[^/]*/?$",
                    3,
                ),
            ]
            self.trim = True

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "vbulletin-3+-thread", "url": url, "id": int(i_id)}
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                    .title {
                        [0] * .threadtitle; * c@[0] m@>[!0] | "%Di" ||
                        [0] * .main-title c@[0] | "%Di" ||
                        td .navbar; strong child@; * [0] c@[0] m@>[!0] | "%Di" / sed "s/<[^>]+>//g" "E" ||
                        * .thread-breadcrumb; * [0] c@[0] m@>[!0] | "%Di" ||
                        * #toolbar; [0] h1 c@[0] | "%Di"
                    } / trim,
                    .path.a {
                        span c@[1:5] .navbar; a [0] m@>[!0] child@ | "%Di\n" ||
                        {
                            div itemtype="http://schema.org/BreadcrumbList" ||
                            [0] * ( #breadcrumbs )( .breadcrumbs ) ||
                            * .navbit
                        }; a c@[:5]; * [0] c@[0] m@>[!0] | "%Di\n"
                    } / trim "\n"
            """
                )
            )
            dict_add(ret, t)

            posts = []
            expr = reliq.expr(
                r"""
                .posts {
                    [0] ul .conversation-list,
                    * #posts; {
                        ul child@ ||
                        * self@
                    }
                }; * -#lastpost -.ad_block child@; {
                    .date {
                        time itemprop=dateCreated | "%(datetime)v" ||
                        span .date | "%DT" ||
                        span .timepost | "%i" ||
                        a name=E>post[0-9]+; * parent@ | "%DT" / sed "s/<[^>]+>//g" "E" trim
                    },

                    .id.u {
                        [0] a name=E>post[0-9]+ | "%(name)v" ||
                        [0] a #E>post[0-9]+ | "%(id)v"
                    } / sed "s/^post//",

                    .title {
                        h2 ( .title )( .b-post__title )( .posttitle ) | "%Di" ||
                        img .inlineimg; div .smallfont c@[:3] parent@ | "%Di" ||
                        div .tittle | "%Di"
                    } / trim,

                    .content div ( #b>post_message_ )( .js-post__content-text ) | "%i",

                    .signature {
                        div .post-signature | "%i" ||
                        div .signaturecontainer | "%i" ||
                        [0] div m@bB>"[^<]*________" | "%Di"
                    },

                    .edited {
                        div .b-post__edit ||
                        div .smallfont; [0] em child@ ||
                        blockquote .lastedited
                    }; {
                        [0] a; {
                            .userid.u * self@ | "%(data-vbnamecard)v",
                            .user_link * self@ | "%(href)v"
                        },
                        .user {
                            [0] a | "%i" ||
                            * self@ | "%t" / tr "\n\t\r" sed "s/[(;)].*//g; s/  *$//; s/.* //"
                        },
                        .date {
                            * self@ | "%t" / tr "\n\t\r" sed "s/.*[(;]//g",
                            span .time | " %i"
                        } / sed "s/)$//" trim
                    },

                    .attachments li .b-post-attachments__item; a [0]; {
                        .link * self@ | "%(href)v",
                        .name span .h-wordwrap | "%i" / sed "s/&lrm;$//" decode trim,
                        span c@[0] m@e>")"; {
                            .size * self@ | "%i" / sed "s/^(//;s/,.*//",
                            .views.u * self@ | "%i" / sed "s/.*,//"
                        }
                    } | ,
                    .attachments2 {
                        div .attachments; li c@[1:] ||
                        a href=Eb>"(attachment|download)\.php\?"; td parent@
                    }; {
                        a [0]; {
                            .link * self@ | "%(href)v",
                            .name * self@ | "%i" / sed "s/&lrm;$//" decode trim
                        },
                        .size * self@ | "%t" / sed "/(/{ s/.*(//; s/,.*//; p }" "n" trim,
                        .views.u * self@ | "%t" / sed "/(/{ s/.*(//; s/.*,//; p }" "n"
                    } | ,


                    .reactions div .reactions__topic-reaction; {
                        .type [0] span .reactions__emoji | "%Di",
                        .count.u [0] span .reactions__topic-reaction--count | "%i"
                    } | ,

                    .thankedby_users.a {
                        div .likes_postbit; li; a [0]; * c@[0] | "%i\n" ||
                        * #b>post_thanks_box_; [-] a; * parent@; * c@[0] | "%i\n"
                    },
                    .thankedby_dates.a(",") * #b>post_thanks_box_; [-] a; * parent@ | "%Dt" / tr "\n\t\r() " tr "[[:print:]]" "" "c" sed "s/<!--[^>]*>//g; s#<[^>]*>##g",

                    .user {
                        [0] * #b>postmenu_; [-] * -C@"[0] * #b>td_post_" ancestor@ ||
                        [0] div ( .userinfo )( .userinfo_noavatar )( .postauthor )
                    }; {
                        {
                            [0] * .username ||
                            [0] a .bigusername ||
                            div .author ||
                            [0] a .siteicon_profile ||
                            * #b>postmenu_ c@[0]
                        }; {
                            .link [0] a | "%(href)v",
                            .name [0] * c@[0] | "%Di" / trim
                        },
                        .avatar {
                            [0] * .E>"(postuseravatar(link)?|(b-)?avatar(-block)?|userpic|user-avatar-h)"; img | "%(src)v" ||
                            [0] img src=Ea>"((/|^)(forum|custom)?avatars?/|image\.php\?)" | "%(src)v"
                        },

                        .title {
                            * ( .usertitle )( .usertittle ) ||
                            * c@[0] m@>[1:]; [0] * c@[:5] .smallfont ancestor@ ||
                            [0] * c@[0] .smallfont
                        }; [0] * c@[0] m@>[1:] | "%Di" / trim,

                        .rank.u {
                            [0] * ( .rank )( .b-userinfo__rank ) | "%c" ||
                            [-] * #b>repdisplay_ | "%c" ||
                            * b-meter__bar m@>[1:] | "l" / wc "c"
                        },
                        .rank-name [0] a .rank-link | "%Di",

                        .custom1 * ( .userinfo_extra )( .userfield ); dl; dt; {
                            .key * self@ | "%Di" trim,
                            .value [0] dd ssub@; [0] * c@[0] m@>[1:] | "%Di" trim
                        } | ,
                        .custom2 li .b-userinfo__additional-info; {
                            .key label child@ | "%Di" trim,
                            .value span child@; [0] * c@[0] m@>[1:]  | "%Di" trim
                        } | ,
                        .custom3 div .fields-list; div .line; {
                            .key span child@ | "%Di" trim,
                            .value strong child@ | "%Di" trim
                        } | ,
                        .custom4.a {
                            div c@[0] m@Bf>" *"; [0] div ( .smallfont )( .userinfo ) parent@ | "%Di" trim ||
                            div a@[0] c@[:2] m@":"; [0] div c@[:14] parent@ | "%Di" trim
                        } / tr "\r\t" sed "
                            :r
                            s/ //g
                            s/<!--[^>]*>//g
                            s/<div>//g
                            s/<\/div>/\n/g
                            s/<b>//g
                            s/<\/b>//g
                            s/< *br *(\/ *)?>/\n/g

                            /^ *$/d
                            /: *$/{
                                N
                                s/\n/ /g
                                br
                            }
                        " "E" sed "/^ *$/d; s/^ *//; s/ *$//"
                    }
                } |
            """
            )

            while True:
                t = json.loads(rq.search(expr))
                posts += t["posts"]

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break

                try:
                    rq, ref = self.session.get_html(
                        nexturl, settings, state, trim=self.trim
                    )
                except AlreadyVisitedError:
                    break

            outposts = []
            for i in posts:
                thankedby = []
                for j in range(len(i["thankedby_users"])):
                    thankedby.append(
                        {
                            "user": i["thankedby_users"][j],
                            "date": (
                                i["thankedby_dates"][j]
                                if j < len(i["thankedby_dates"])
                                else ""
                            ),
                        }
                    )
                i.pop("thankedby_users")
                i.pop("thankedby_dates")
                i["thankedby"] = thankedby

                if i["date"] == "" and i["id"] == 0 and i["content"] == "":
                    if len(outposts) > 0:
                        outposts[len(outposts) - 1]["thankedby"] += thankedby
                    continue

                i["edited"]["user_link"] = url_merge_r(ref, i["edited"]["user_link"])

                attachments = i["attachments"]
                if len(attachments) < len(i["attachments2"]):
                    attachments = i["attachments2"]
                i.pop("attachments2")
                for j in attachments:
                    j["link"] = url_merge_r(ref, j["link"])
                i["attachments"] = attachments

                user = i["user"]
                user["link"] = url_merge_r(ref, user["link"])
                user["avatar"] = url_merge_r(ref, user["avatar"])

                custom = user["custom1"]
                if len(custom) < len(user["custom2"]):
                    custom = user["custom2"]
                if len(custom) < len(user["custom3"]):
                    custom = user["custom3"]
                if len(custom) < len(user["custom4"]):
                    custom = []
                    for j in user["custom4"]:
                        tokens = j.split(":", 1)
                        if len(tokens) < 2:
                            custom.append({"key": None, "value": j})
                        else:
                            custom.append(
                                {"key": tokens[0].strip(), "value": tokens[1].strip()}
                            )
                user.pop("custom1")
                user.pop("custom2")
                user.pop("custom3")
                user.pop("custom4")
                user["custom"] = custom

                outposts.append(i)

            ret["posts"] = outposts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_vbulletin

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.trim = True

        self.forum_forums_expr = reliq.expr(
            r'''{
                * .forumtitle; a l@[:2] c@[0] | "%(href)v\n",
                * #E>f[0-9]+; a [0] c@[:4] | "%(href)v\n",
                a .forum-title c@[0] | "%(href)v\n",
                a href=E>"/f[0-9]+/" c@[0] | "%(href)v\n",
                h2 .forumTitle; a [0] l@[:3] c@[:3] | "%(href)v\n"
            } / sed "s/&amp;/\&/g" sort "u"'''
        )
        self.forum_threads_expr = reliq.expr(
            r'a ( #b>thread_title_ )( .topic-title ) href | "%(href)v\n" / sed "s/&amp;/\&/g"'
        )
        self.board_forums_expr = self.forum_forums_expr
        self.guesslist = [
            {
                "func": "get_thread",
                "exprs": [
                    r"/showthread\.php\?",
                ],
            },
            {
                "func": "get_forum",
                "exprs": [r"/forumdisplay\.php(\?|/)", "/forums?/[^/]+/?$"],
            },
            {
                "func": "get_thread",
                "exprs": [
                    r"(/[^/]+){2,}/\d+-[^/]+$",
                    r"(/[^/]+){2,}/[^/]+-\d+(\.html|/)$",
                ],
            },
            {"func": "get_forum", "exprs": [r"(/[^/]+){3,}/$"]},
            {
                "func": "get_board",
                "exprs": [r"/forums?(/|\.php(\?[^/]*)?)$"],
            },
            {"func": "get_board", "exprs": None},
        ]

        self.findroot_expr = reliq.expr(
            r"""
                {
                    [0] * ( #breadcrumbs )( .breadcrumbs )( #breadcrumb )( .breadcrumbs-table ); a href=>[1:] | "%(href)v\n" / sed "
                        \#(/|^)(((forum|foro|board)s?|community)/?|forum\.php)(\?[^/]*)?$#{p;q}
                        1!G
                        h
                       $p" "En" line [-],
                    [0] span .navbar; a href | "%(href)v\n"
                } / line [0] tr "\n" sed "s/&amp;/\&/g"
            """
        )
        self.findroot_board = True
        self.findroot_board_expr = re.compile(
            r"^(/[^\.-])?/((board|forum|foro)s?|(index|forum)\.(php|html)|community|communaute|comunidad)(/|\?[^/]*)?$",
        )

    def get_next_page(self, rq):
        url = rq.search(
            r'''{
                * ( .pagenav )( .threadpagenav )( .pagination )( #mb_pagenav ); {
                    [0] a rel=next | "%(href)v",
                    [0] a .js-pagenav-next-button href=b>http | "%(href)v",
                },
                span .pages; [0] span .currentpage; [0] a ssub@ | "%(href)v"
            } / line [0] sed "/^javascript:/d; s/&amp;/\&/g"'''
        )

        return url

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                .categories {
                    * #b>collapseobj_forumbit_ ||
                    * ( #E>forum[0-9]+ )( .subforum-list ) ||
                    * #E>f[0-9]+; * parent@ ||
                    li .forumRow
                }; {
                    .header * self@ #b>collapseobj_forumbit_; {
                        [0] * spre@; * -#b>collapseobj_forumbit_ self@ ||
                        * parent@; [0] * spre@; * -#b>collapseobj_forumbit_ self@
                    }; {
                        .name {
                            [0] a -C@"[0] img"; [0] * c@[0] m@>[1:] | "%Di" ||
                            [0] a; * parent@ | "%Dt"
                        } / trim,
                        .link [0] a -C@"[0] img" | "%(href)v",
                        .description [0] * .smallfont c@[0] m@>[1:] | "%Di",
                    },
                    .header2 * self@ #E>forum[0-9]+; {
                        * .category-header self@ ||
                        [1] * ancestor@; * #E>cat[0-9]+ self@
                    }; {
                        .name {
                            a .category c@[0] | "%Di" ||
                            * .catTitle; [0] a; * c@[0] m@>[1:] | "%Di" ||
                            [0] * .forumhead; span .forumtitle; [0] a; * c@[0] m@>[1:] | "%Di"
                        } / trim,
                        .link {
                            a .category c@[0] | "%(href)v" ||
                            * .catTitle; [0] a | "%(href)v" ||
                            [0] * .forumhead; * .forumtitle; [0] a | "%(href)v"
                        },
                        .description [0] * .forumhead; * .subforumdescription; [0] * c@[0] m@>[1:] | "%Di"
                    },

                    .forums {
                        * self@ #b>collapseobj_forumbit_; * -thead child@ ||
                        * self@
                    }; {
                        .status {
                            [0] img ( #b>forum_statusicon_ )( .forumicon ) | "%(src)v" ||
                            [0] span .icon | "%(title)v"
                        },
                        {
                            * .forum-title ||
                            [0] v>span .forumtitle ||
                            [0] a c@[:1]
                        }; {
                            .title [0] * c@[0] m@>[1:] | "%Di" trim,
                            .link [0] a | "%(href)v"
                        },
                        .description {
                            [0] * ( .forum-desc )( .forumdescription )( .description ) | "%i" ||
                            div .smallfont; {
                                [0] * c@[0] m@b>[1:] self@ | "%i" ||
                                [0] * child@; v>strong * self@; * parent@ | "%i"
                            }
                        },
                        .viewing.u {
                            span .viewing | "%i" ||
                            [0] span c@[0] m@tb>"(" m@te>")" | "%i"
                        },

                        .childboards {
                            * .subforum-info ||
                            [0] * .subforums; * child@ ||
                            strong c@[0]; div .smallfont parent@
                        }; a; {
                            .link * self@ | "%(href)v",
                            .name [0] * c@[0] m@>[1:] | "%Di" / trim,
                            .icon [0] * spre@; {
                                img self@ | "%(src)v" ||
                                span self@ | "%(title)v"
                            },
                            [0] * ssub@; span c@[0] m@tb>"(" m@te>")"; {
                                .topics.u * self@ | "%i",
                                .posts.u * self@ | "%i" / sed "/\//!d; s#.*/##"
                            }
                        } | ,

                        .topics.u {
                            [0] * ( .topics-count )( .threadcount ) c@[0] | "%i" ||
                            * ( .forumstats )( .forumstats_2 ); [0] * c@[0] child@ | "%i" ||
                            [0] * c@[0] m@tEf>"(([0-9]+,)+)?[0-9]+" | "%i" ||
                            [0] * c@[0] m@tEf>"(([0-9]+,)+)?[0-9]+ */ *(([0-9]+,)+)?[0-9]+" | "%i"
                        } / tr ",.",
                        .posts.u {
                            [0] * ( .posts-count )( .postcount ) c@[0] | "%i" ||
                            * ( .forumstats )( .forumstats_2 ); [1] * c@[0] child@ | "%i" ||
                            [1] * c@[0] m@tEf>"(([0-9]+,)+)?[0-9]+" | "%i" ||
                            [0] * c@[0] m@tEf>"(([0-9]+,)+)?[0-9]+ */ *(([0-9]+,)+)?[0-9]+" | "%i" / sed "s/.*\///"
                        } / tr ",.",

                        .lastpost {
                            [0] * ( .i>lastpost )( .i>forumlastpost ) ||
                            strong c@[0] m@>[1:]; a parent@; div [0] l@[0:5] .smallfont ancestor@; [-] * self@
                        }; {
                            {
                                [0] * ( .lastpost-title )( .lastposttitle )( .title ) ||
                                [0] strong c@[0]; * parent@
                            }; {
                                .title [0] * c@[0] m@>[1:] | "%Di" / trim,
                                .link [0] a | "%(href)v",
                            },
                            .icon {
                                * .lastposttitle; [0] img | "%(src)v" ||
                                [0] img .inlineimg | "%(src)v"
                            },
                            .date {
                                [0] * ( .lastpost-date )( .lastpostdate )( .time ) | "%DT" trim,
                                [-] div; * self@ c@[:5] | "%DT" trim
                            },
                            .user {
                                * ( .lastpost-by )( .lastpostby ); [0] a ||
                                * .title ||
                                [-1] div; * c@[:1] self@
                            }; [0] * c@[0] m@>[1:] | "%Di" trim,
                            .user_link {
                                * ( .lastpost-by )( .lastpostby ) ||
                                [-1] div; * c@[1] self@
                            }; [0] a; * self@ | "%(href)v",
                            .avatar a .avatar; [0] img | "%(src)v"
                        }
                    } |
                } | ,

                .threads {
                    li #b>thread_ ||
                    * .topic-item ||
                    * #b>threadbits_forum_; * child@ ||
                    * #threadslist; * child@
                }; {
                    .avatar a .avatar; [0] img | "%(src)v",
                    .icons.a {
                        [0] * .threadtitle; img child@ | "%(src)v\n" ||
                        img .js-post-icon | "%(src)v\n" ||
                        {
                            img #b>thread_statusicon_ | "%(src)v\n",
                            div c@[1]; img -id | "%(src)v\n",
                            * #b>td_threadtitle_; img | "%(src)v\n"
                        }
                    },
                    .prefix {
                        [0] * ( .topic-prefix )( .js-topic-prefix )( .prefix )( #b>thread_prefix_ ) | "%DT" trim ||
                        span c@[0] m@et>":" | "%Di"
                    },
                    {
                        [0] * ( .topic-title )( #b>thread_title_ ) ||
                        [0] * #b>thread_title_
                    }; [0] a; {
                        .link * self@ | "%(href)v",
                        .title * c@[0] m@>[1:] | "%Di" / trim
                    },
                    {
                        div .topic-info ||
                        div .threadmeta ||
                        * #b>td_threadtitle_; [-] div .smallfont
                    }; {
                        .user {
                            div .smallfont c@[:4] self@ | "%DT" trim ||
                            span style=b>cursor:; [0] * c@[0] m@>[1:] | "%Di" trim ||
                            [0] a; [0] * c@[0] m@>[1:] | "%Di" trim ||
                        },
                        .user_link {
                            [0] span onclick | "%(onclick)v" / sed "s/,.*//;s/.*'([^']+)'/\1/" "E" ||
                            [0] a | "%(href)v"
                        },
                        .date {
                            [0] span ( .date )( .time ) | "%Di" trim ||
                            span style=b>cursor:; [0] * ssub@; span self@; [0] * c@[0] m@>[1:] | "%Di" trim ||
                            div .author; [0] span c@[0] child@ m@E>" [0-9]{4}" | "%Di" trim ||
                            [0] a; * parent@ | "%Dt" / sed "s/.*,//; s/ //g" trim
                        } / sed "s/^on //"
                    },
                    .detailicons.a {
                        * .threaddetailicons; img | "%(src)v\n" ||
                        * .cell-icons; span | "%(title)v\n"
                    },
                    .posts.u {
                        {
                            div .posts-count ||
                            * .threadstats; {
                                [0] li -.hidden; [0] * c@[0] m@>[1:] ||
                                * self@
                            }
                        }; * self@ | "%i" ||
                        * c@[4:] title=aE>"[0-9]+.*,.*[0-9]+" | "%(title)v" / sed "s/, .*//"
                    } / tr ",.",
                    .views.u {
                        div .views-count | "%i" ||
                        * .threadstats; {
                            [1] li -.hidden; [0] * c@[0] m@>[1:] | "%i" ||
                            * self@ | "%i" / sed "s/.*>//"
                        }  ||
                        * c@[4:] title=aE>"[0-9]+.*,.*[0-9]+" | "%(title)v" / sed "s/.*, //"
                    } / tr ",.",
                    .reactions.u div .votes-count | "%i" tr ",.",
                    .rating.u {
                        [0] * .E>rating[0-9] | "%(class)v" ||
                        [0] img src=eE>"/rating_?[0-9]+.gif" | "%(src)v" / sed "s#.*/##"
                    },
                    .lastpage.u {
                        * #b>pagination_threadbit_; [-] a | "%i" ||
                        [0] * m@tb>"(" m@te>")" c@[1:]; [-] a; * self@ | "%(href)v" / sed "s#/$##; s#.*/##; s/.*;page=//" ||
                        * #b>td_threadtitle_; [-] a m@Eft>"[0-9]+" | "%i"
                    },

                    .lastpost [0] * ( .cell-lastpost )( .threadlastpost ); {
                        .avatar [0] a .avatar; [0] img src | "%(src)v",
                        {
                            * .lastpost-by; [0] a ||
                            [0] a .username
                        }; {
                            .user_link * self@ | "%(href)v",
                            .user * c@[0] m@>[1:] | "%Di"
                        },
                        .date {
                            span .post-date | "%Di" ||
                            * .time; * parent@ | "%Dt" trim ||
                            [0] * c@[0] m@E>"[0-9]{4}" | "%Di" / sed "s/ //g" trim
                        },
                        .link [0] a ( .go-to-last-post )( .lastpostdate ) | "%(href)v"
                    },
                    .lastpost_link [0] a .lastpostdate | "%(href)v",
                    .lastpost2 * c@[3:] title=aE>"[0-9]+.*,.*[0-9]+"; div .smallfont child@; {
                        .avatar nothing | "",
                        .user_link [0] a c@[0] | "%(href)v",
                        .user {
                            [0] a c@[0] | "%Di" trim ||
                            * self@ | "%Di" / tr "\n" sed "s/.*<br ?\/?>//" "E" trim
                        },
                        .date * self@ | "%i" / tr "\n" sed "
                            s/<br ?\/?>.*//
                            s/<[^>]*>//g
                        " "E" trim,
                        .link img; [-] a parent@ | "%(href)v"
                    }
                } |
                """
            )
        )

        categories = []
        prevcat = None
        for i in t["categories"]:
            c = {}
            header = i["header"]
            if header["name"] == "" and header["link"] == "":
                header = i["header2"]
            i.pop("header")
            i.pop("header2")

            header["link"] = url_merge(ref, header["link"])
            c["name"] = header["name"]
            c["link"] = header["link"]
            c["description"] = header["description"]

            forums = []
            for j in i["forums"]:
                if j["title"] == "" and j["link"] == "":
                    continue

                if j["status"].find("/"):
                    j["status"] = url_merge(ref, j["status"])

                j["link"] = url_merge(ref, j["link"])

                for k in j["childboards"]:
                    k["link"] = url_merge(ref, k["link"])
                    if k["icon"].find("/"):
                        k["icon"] = url_merge(ref, k["icon"])

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["icon"] = url_merge(ref, lastpost["icon"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
                lastpost["avatar"] = url_merge(ref, lastpost["avatar"])

                forums.append(j)

            if prevcat is not None and c["link"] == prevcat["link"]:
                prevcat["forums"] += forums
            elif (
                prevcat is not None
                and c["link"] is None
                and (
                    len(prevcat["forums"]) == 0
                    or (
                        len(prevcat["forums"]) == 1
                        and prevcat["forums"][0]["link"] == prevcat["link"]
                    )
                )
            ):
                prevcat["forums"] = forums
            else:
                c["forums"] = forums
                prevcat = c
                categories.append(c)

        threads = []
        for i in t["threads"]:
            if i["link"] == "" and i["title"] == "" and i["user"] == "":
                continue

            i["avatar"] = url_merge(ref, i["avatar"])
            for j, icon in enumerate(i["icons"]):
                i["icons"][j] = url_merge(ref, i["icons"][j])

            i["link"] = url_merge(ref, i["link"])
            i["user_link"] = url_merge(ref, i["user_link"])
            for j, icon in enumerate(i["detailicons"]):
                if icon.find("/") != -1:
                    i["detailicons"][j] = url_merge(ref, icon)

            lastpost = i["lastpost"]
            if (
                lastpost["user"] == ""
                and lastpost["user_link"] == ""
                and lastpost["date"] == ""
            ):
                lastpost = i["lastpost2"]
            if lastpost["link"] == "":
                lastpost["link"] = i["lastpost_link"]
            i.pop("lastpost")
            i.pop("lastpost2")
            i.pop("lastpost_link")

            lastpost["link"] = url_merge(ref, lastpost["link"])
            lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
            lastpost["avatar"] = url_merge(ref, lastpost["avatar"])

            i["lastpost"] = lastpost

            threads.append(i)

        return {
            "format_version": "vbulletin-3+-forum",
            "url": url,
            "categories": categories,
            "threads": threads,
        }



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..enums import Outputs
from ..utils import dict_add, get_settings, url_merge_r, conv_short_size, url_merge
from .common import ItemExtractor, ForumExtractor
from .identify import identify_hackernews


def get_comments(ref, rq):
    comments = json.loads(
        rq.search(
            r"""
        .comments tr id .athing .comtr; {
            .id.u * l@[0] | "%(id)v",
            table; tr; {
                .lvl.u td .ind indent | "%(indent)v",
                td .default; {
                    span .comhead;  {
                        .user a .hnuser href; {
                            .link * l@[0] | "https://news.ycombinator.com/%(href)v",
                            .name [0] * c@[0] m@>[1:] | "%Di" trim
                        },
                        .date span .age title | "%(title)v",
                        .onstory span .onstory; [0] a; {
                            .name * self@ | "%Di" / trim,
                            .link * self@ | "%(href)v"
                        }
                    },
                    .text div .commtext | "%i"
                }
            }
        } |
    """
        )
    )["comments"]

    for i in comments:
        i["user"]["link"] = url_merge(ref, i["user"]["link"])
        i["onstory"]["link"] = url_merge(ref, i["onstory"]["link"])
    return comments


def get_post(ref, rq):
    post = json.loads(
        rq.search(
            r"""
        tr id .athing l@[:1]; {
            .id.u * self@ | "%(id)v",
            span .titleline; [0] a; {
                .link * self@ href | "%(href)v",
                .title * self@ | "%Di" trim
            }
        },
        tr -class l@[:1]; {
            .score.u span #b>score_ | "%i" sed "s/ .*//",
            .date span .age title | "%(title)v",
            .user a .hnuser href; {
                .link * self@ | "%(href)v",
                .name [0] * c@[0] m@>[1:] | "%Di" trim
            },
            a m@"&nbsp;comment"; {
                .comments_count.u a self@ | "%i" sed "s/&.*//",
                .comments_link a self@ | "%(href)v"
            }
        },
        .text [0] div .toptext | "%i"
        """
        )
    )

    post["link"] = url_merge(ref, post["link"])
    post["user"]["link"] = url_merge(ref, post["user"]["link"])
    post["comments_link"] = url_merge(ref, post["comments_link"])
    return post


def get_page(ref, rq):
    threads = []
    posts_list = rq.search(r'[1] table -#hnmain; tr l@[1] | "%C\0"').split("\0")[:-1]
    size = len(posts_list)
    i = 0
    while i < size and size - i >= 3:
        inp = posts_list[i] + posts_list[i + 1] + posts_list[i + 2]

        post = get_post(ref, reliq(inp))
        threads.append(post)

        i += 3

    return threads


def get_all_pages(self, ref, rq, settings, state):
    threads = []
    page = 0

    while True:
        threads += get_page(ref, rq)

        page += 1
        if settings["thread_pages_max"] != 0 and page >= settings["thread_pages_max"]:
            break
        url = self.get_next(ref, rq)
        if url is None:
            break

        rq, ref = self.session.get_html(url, settings, state, True)

    return threads


def get_all_comments(self, ref, rq, settings, state):
    comments = []
    page = 0

    while True:
        comments += get_comments(ref, rq)

        page += 1
        if settings["thread_pages_max"] != 0 and page >= settings["thread_pages_max"]:
            break
        url = self.get_next(ref, rq)
        if url is None:
            break

        rq, ref = self.session.get_html(url, settings, state, True)

    return comments


class hackernews(ForumExtractor):
    class User(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"^https://news.ycombinator.com/user\?id=([^&]+)"),
                    1,
                )
            ]
            self.path_format = "m-{}"
            self.trim = True

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "hackernews-user", "url": url, "id": i_id}

            t = json.loads(
                rq.search(
                    r"""
                    table #hnmain; [1] table desc@; tr; {
                        [0] td m@f>"user:"; td ssub@; {
                            .created-timestamp * timestamp self@ | "%(timestamp)v",
                            .user [0] a; [0] * c@[0] m@>[1:] | "%Di" trim
                        },
                        .created-date [0] td m@f>"created:"; td ssub@; [0] a | "%i",
                        .karma.u [0] td m@f>"karma:"; td ssub@ | "%i",
                        .about [0] td m@f>"about:"; td ssub@ | "%i",
                        .submissions-link [0] a href=b>"submitted?" | "%(href)v",
                        .comments-link [0] a href=b>"threads?" | "%(href)v",
                        .favorites-link [0] a href=b>"favorites?" | "%(href)v",
                    }
                    """
                )
            )

            t["submissions-link"] = url_merge_r(ref, t["submissions-link"])
            t["comments-link"] = url_merge_r(ref, t["comments-link"])
            t["favorites-link"] = url_merge_r(ref, t["favorites-link"])

            rq, ref = self.session.get_html(
                t["submissions-link"], settings, state, True
            )
            t["submissions"] = get_all_pages(self, ref, rq, settings, state)

            rq, ref = self.session.get_html(t["comments-link"], settings, state, True)
            t["comments"] = get_all_comments(self, ref, rq, settings, state)

            rq, ref = self.session.get_html(t["favorites-link"], settings, state, True)
            t["favorites"] = get_all_pages(self, ref, rq, settings, state)

            dict_add(ret, t)
            return ret

    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"^https://news.ycombinator.com/item\?id=(\d+)"),
                    1,
                )
            ]
            self.trim = True

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "hackernews-thread", "url": url, "id": i_id}

            fatitem = rq.filter(r"[0] table .fatitem")
            t = get_post(ref, fatitem)
            dict_add(ret, t)

            ret["comments"] = get_all_comments(self, ref, rq, settings, state)
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_hackernews

        self.trim = True

        self.thread = self.Thread(self.session)
        self.user = self.User(self.session)
        self.user.get_next = self.get_next
        self.thread.user = self.user
        self.thread.get_next = self.get_next

        self.domains = ["news.ycombinator.com"]
        self.domain_guess_mandatory = True

        self.forum_threads_expr = reliq.expr(
            r'span .subline; a [0] m@"&nbsp;comment" href | "%(href)v\n"'
        )
        self.guesslist = [
            {
                "func": "get_thread",
                "exprs": [r"/item\?id="],
            },
            {
                "func": "get_user",
                "exprs": [r"/user\?id="],
            },
            {
                "func": "get_forum",
                "exprs": [
                    r"/(news|newest|front|show(new)?|ask|jobs)($|\?p=)",
                    r"/(favorites|submitted)\?id=",
                ],
            },
        ]

    def get_next_page(self, rq):
        return rq.search(r'[0] a .morelink rel=next href | "%(href)v"')

    def process_forum_r(self, url, ref, rq, settings, state):
        threads = get_page(ref, rq)

        return {
            "format_version": "hackernews-forum",
            "url": url,
            "threads": threads,
        }



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..utils import (
    dict_add,
    get_settings,
    url_merge_r,
    url_merge,
    url_valid,
    conv_short_size,
)
from .common import ItemExtractor, ForumExtractor
from .identify import identify_stackexchange


class stackexchange(ForumExtractor):
    class User(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/users/(\d+)"),
                    1,
                )
            ]
            self.path_format = "m-{}"

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "stackexchange-user", "url": url, "id": int(i_id)}

            t = json.loads(
                rq.search(
                    r"""
                div #mainbar-full; {
                    [0] div child@; {
                        .avatar a; [0] img src | "%(src)v",
                        .user [0] div .fs-headline2 c@[0] | "%Di" trim,
                        .unregistered.b div .s-badge c@[0] | "t",
                        ul .list-reset; {
                            [0] * self@; li child@; {
                                .created [0] * self@; span title | "%(title)v",
                                .lastseen [1] * self@; div c@[0] m@bt>"Last seen" | "%i" sed "s/^Last seen//" trim
                            },
                            [1] * self@; {
                                .location div .wmx2 title c@[0] | "%i",
                                .social.a a href | "%(href)v\n"
                            }
                        },
                        li role=menuitem; {
                            .meta-profile a href=a>meta. | "%(href)v",
                            .network-profile a href=b>https://stackexchange.com/users/ | "%(href)v"
                        },
                    },
                    main #main-content; {
                        div #stats; div .md:fl-auto; {
                            .reputation * self@ m@reputation; div child@ c@[0] | "%i" tr ",",
                            .reached * self@ m@reached; div child@ c@[0] | "%i" tr ",",
                            .answers * self@ m@answers; div child@ c@[0] | "%i" tr ",",
                            .questions * self@ m@questions; div child@ c@[0] | "%i" tr ","
                        },
                        div c@[0] m@tf>Communities; [1] * ancestor@; {
                            .communities-all [0] a m@bt>"View all" | "%(href)v",
                            .communities li; a; {
                                .profile * self@ | "%(href)v",
                                .reputation div .ml-auto c@[0] | "%i" tr ","
                            } |
                        },
                        .about [0] div .js-about-me-content | "%i",
                        div c@[0] m@tf>Badges; [1] * ancestor@; {
                            .badges-all [0] a m@bt>"View all" | "%(href)v",
                            .badges div .s-card; div .jc-space-between c@[6:] child@; {
                                .amount.u div .fs-title | "%i",
                                .name div .fs-caption | "%Di" line [0] " " tr " " / trim,
                                .achievements li; {
                                    a ( .badge )( .badge-tag ); {
                                        .tag.b * self@ .badge-tag | "t",
                                        .link * self@ | "%(href)v",
                                        .action * self@ | "%(title)v" / sed 's/^[^:]*: //',
                                        .name div c@[0] | "%Di" / trim
                                    },
                                    .amount.u div -.ml-auto child@ | "%t",
                                    .date div .ml-auto child@ c@[0] | "%i"
                                } |
                            } |
                        },
                        div c@[0] .fs-title m@tf>"Top tags"; [1] * ancestor@; {
                           .tags-all [0] a m@bt>"View all" | "%(href)v",
                           .tags [0] div .s-card; div child@; {
                                a .s-tag; {
                                    .link * self@ | "%(href)v",
                                    .name * self@ | "%Di" / trim
                                },
                                .badge [0] a .badge-tag title | "%(title)v" / sed "s/ .*//",
                                div .d-flex .ai-center; {
                                    .score div .tt-lowercase m@tf>Score; [0] * spre@ | "%i" tr ",",
                                    .posts div .tt-lowercase m@tf>Posts; [0] * spre@ | "%i" tr ",",
                                    .posts-percent.u div .tt-lowercase m@tf>"Posts %"; [0] * spre@ | "%i" tr ",",
                                },
                                .answered.u * self@ title | "%(title)v" / sed "/Gave/!d; s/.*. Gave ([0-9]+) non-wiki .*/\1/" "E",
                                .ansked.u * self@ title | "%(title)v" / sed "s/^Asked ([0-9]+) .*/\1/" "E",
                                .asked-score.u * self@ title | "%(title)v" / sed "/^Asked/!d; s/\..*//; s/^.*total score //" "E"
                            } |
                        },
                        div c@[0] .fs-title m@tf>"Top posts"; [2] * ancestor@; {
                            div m@bt>"View all"; a; {
                                .posts-answers-all * self@ m@ft>answers | "%(href)v",
                                .posts-questions-all * self@ m@ft>questions | "%(href)v"
                            },
                            .posts [0] div .s-card; div child@; {
                                .type [0] title | "%i",
                                .answered.b [0] div .s-badge__answered | "t",
                                .votes.u div .s-badge__votes | "%i",
                                [0] a .d-table; {
                                    .link * self@ | "%(href)v",
                                    .title * self@ | "%Di" trim
                                },
                                .date span .relativetime title | "%(title)v"
                            } |
                        },
                        div c@[0] .fs-title m@tf>"Top network posts"; [1] * ancestor@; {
                           .network-posts-all [0] a m@bt>"View all" | "%(href)v",
                           .network-posts div .s-card; div child@; {
                                .votes.u div .s-badge__votes | "%i",
                                a .d-table; {
                                    .link * self@ | "%(href)v",
                                    .title * self@ | "%Di" trim
                                },
                            } |
                        },
                        div c@[0] .fs-title m@tf>"Top Meta posts"; [1] * ancestor@; {
                            .meta-posts-asked.u div .ml8 title=b>asked | "%(title)v",
                            .meta-posts-answered.u div .ml8 title=b>gave | "%(title)v",
                            .meta-posts div .s-card; div child@; {
                                .votes.u div .s-badge__votes | "%i",
                                a .d-table; {
                                    .link * self@ | "%(href)v",
                                    .title * self@ | "%Di" trim
                                },
                            } |
                        }
                    }
                }
            """
                )
            )
            t["avatar"] = url_merge_r(ref, t["avatar"])
            t["meta-profile"] = url_merge_r(ref, t["meta-profile"])
            t["network-profile"] = url_merge_r(ref, t["network-profile"])

            t["reputation"] = conv_short_size(t["reputation"])
            t["reached"] = conv_short_size(t["reached"])
            t["answers"] = conv_short_size(t["answers"])
            t["questions"] = conv_short_size(t["questions"])

            t["communities-all"] = url_merge_r(ref, t["communities-all"])
            for i in t["communities"]:
                i["profile"] = url_merge_r(ref, i["profile"])
                i["reputation"] = conv_short_size(i["reputation"])

            t["badges-all"] = url_merge_r(ref, t["badges-all"])
            for i in t["badges"]:
                for j in i["achievements"]:
                    j["link"] = url_merge_r(ref, j["link"])
                    if j["amount"] == 0:
                        j["amount"] = 1

            t["tags-all"] = url_merge_r(ref, t["tags-all"])
            for i in t["tags"]:
                i["link"] = url_merge_r(ref, i["link"])
                i["score"] = conv_short_size(i["score"])
                i["posts"] = conv_short_size(i["posts"])

            t["posts-answers-all"] = url_merge_r(ref, t["posts-answers-all"])
            t["posts-questions-all"] = url_merge_r(ref, t["posts-questions-all"])
            for i in t["posts"]:
                i["link"] = url_merge_r(ref, i["link"])

            t["network-posts-all"] = url_merge_r(ref, t["network-posts-all"])
            for i in t["network-posts"]:
                i["link"] = url_merge_r(ref, i["link"])

            for i in t["meta-posts"]:
                i["link"] = url_merge_r(ref, i["link"])

            dict_add(ret, t)
            return ret

    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    r"/questions/(\d+)",
                    1,
                )
            ]
            self.trim = True

        def get_post_comments(self, rq, url, ref, postid, settings, state):
            n = rq.search(r'div #b>comments-link-; a .comments-link m@b>"Show " | "t"')
            if len(n) > 0:
                nsettings = get_settings(
                    settings, headers={"x-Requested-With": "XMLHttpRequest"}
                )
                rq, ref = self.session.get_html(
                    "{}/posts/{}/comments".format(url_valid(url, base=True)[0], postid),
                    nsettings,
                    state,
                )

            comments = json.loads(
                rq.search(
                    r"""
                .comments li #E>comment-[0-9]+; {
                    .id.u * self@ | "%(data-comment-id)v",
                    .score.i div .comment-score; * c@[0] | "%i",
                    .content span .comment-copy | "%i",
                    .date span .comment-date; span title | "%(title)v" / sed "s/, L.*//",
                    [0] a .comment-user; {
                        .user * self@ | "%Di" trim,
                        .user_link * self@ | "%(href)v",
                        .reputation.u * self@ | "%(title)v" / tr ",."
                    },
                } |
             """
                )
            )["comments"]

            for i in comments:
                i["user_link"] = url_merge(ref, i["user_link"])
            return comments

        def get_post(self, rq, url, ref, settings, state):
            post = json.loads(
                rq.search(
                    r"""
                .id.u * self@ | "%(data-answerid)v %(data-questionid)v",
                .rating.i div .js-vote-count data-value | "%(data-value)v",
                .checkmark.b [0] div .js-accepted-answer-indicator | "t",
                .bounty.u [0] div .js-bounty-award | "%i",
                .content div class="s-prose js-post-body" | "%i",

                div .user-info m@"edited"; {
                    .edited span .relativetime | "%(title)v",
                    .editor {
                        .avatar img .bar-sm src | "%(src)v",
                         div .user-details; {
                            .name [0] * l@[1] c@[0] | "%Di" trim,
                            .link div .user-details; a href child@ | "%(href)v",
                            .reputation.u span .reputation-score | "%i" tr ",",
                            .gbadge.u span title=w>gold; span .badgecount | "%i",
                            .sbadge.u span title=w>silver; span .badgecount | "%i",
                            .bbadge.u span title=w>bronze; span .badgecount | "%i"
                        }
                    }
                },

                div .user-info m@v>"edited"; {
                    .date span .relativetime | "%(title)v",
                    .author {
                        .avatar img .bar-sm src | "%(src)v",
                         div .user-details; {
                            .name [0] * l@[1] c@[0] | "%Di" trim,
                            .link div .user-details; a href child@ | "%(href)v",
                            .reputation.u span .reputation-score | "%(title)v %i" tr ",",
                            .gbadge.u span title=w>gold; span .badgecount | "%i",
                            .sbadge.u span title=w>silver; span .badgecount | "%i",
                            .bbadge.u span title=w>bronze; span .badgecount | "%i"
                        }
                    }
                },
                """
                )
            )
            post["author"]["avatar"] = url_merge(ref, post["author"]["avatar"])
            post["author"]["link"] = url_merge(ref, post["author"]["link"])
            post["editor"]["avatar"] = url_merge(ref, post["editor"]["avatar"])
            post["editor"]["link"] = url_merge(ref, post["editor"]["link"])

            post["comments"] = self.get_post_comments(
                rq, url, ref, post["id"], settings, state
            )
            return post

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {
                "format_version": "stackexchange-thread",
                "url": url,
                "id": int(i_id),
            }
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                    .title h1 itemprop="name"; a | "%Di" / trim,
                    div .flex--item .mb8 .ws-nowrap; {
                        .views.u [0] * self@ m@"Viewed" | "%(title)v" / tr "0-9" "" "c",
                        .asked [0] * self@ m@"Asked"; [0] time itemprop="dateCreated" datetime | "%(datetime)v",
                        .modified [0] * self@ m@"Modified"; [0] a title | "%(title)v"
                    },
                    .tags.a div .ps-relative; a .post-tag | "%i\n"
                    """
                )
            )
            dict_add(ret, t)

            posts = []
            posts.append(
                self.get_post(
                    rq.filter("div #question").self()[0], url, ref, settings, state
                )
            )

            while True:
                for i in rq.filter(r"div #b>answer-").self():
                    posts.append(self.get_post(i, url, ref, settings, state))

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state)

            ret["posts"] = posts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_stackexchange

        # ucurl 'https://stackexchange.com/sites#' | reliq '[0] div .grid-view-container; a child@ | "%(href)v\n"' | sed 's/^https:\/\//"/;s/$/",/'
        self.domains = [
            "stackoverflow.com",
            "serverfault.com",
            "superuser.com",
            "meta.stackexchange.com",
            "webapps.stackexchange.com",
            "gaming.stackexchange.com",
            "webmasters.stackexchange.com",
            "cooking.stackexchange.com",
            "gamedev.stackexchange.com",
            "photo.stackexchange.com",
            "stats.stackexchange.com",
            "math.stackexchange.com",
            "diy.stackexchange.com",
            "gis.stackexchange.com",
            "tex.stackexchange.com",
            "askubuntu.com",
            "money.stackexchange.com",
            "english.stackexchange.com",
            "stackapps.com",
            "ux.stackexchange.com",
            "unix.stackexchange.com",
            "wordpress.stackexchange.com",
            "cstheory.stackexchange.com",
            "apple.stackexchange.com",
            "rpg.stackexchange.com",
            "bicycles.stackexchange.com",
            "softwareengineering.stackexchange.com",
            "electronics.stackexchange.com",
            "android.stackexchange.com",
            "boardgames.stackexchange.com",
            "physics.stackexchange.com",
            "homebrew.stackexchange.com",
            "security.stackexchange.com",
            "writing.stackexchange.com",
            "video.stackexchange.com",
            "graphicdesign.stackexchange.com",
            "dba.stackexchange.com",
            "scifi.stackexchange.com",
            "codereview.stackexchange.com",
            "codegolf.stackexchange.com",
            "quant.stackexchange.com",
            "pm.stackexchange.com",
            "skeptics.stackexchange.com",
            "fitness.stackexchange.com",
            "drupal.stackexchange.com",
            "mechanics.stackexchange.com",
            "parenting.stackexchange.com",
            "sharepoint.stackexchange.com",
            "music.stackexchange.com",
            "sqa.stackexchange.com",
            "judaism.stackexchange.com",
            "german.stackexchange.com",
            "japanese.stackexchange.com",
            "philosophy.stackexchange.com",
            "gardening.stackexchange.com",
            "travel.stackexchange.com",
            "crypto.stackexchange.com",
            "dsp.stackexchange.com",
            "french.stackexchange.com",
            "christianity.stackexchange.com",
            "bitcoin.stackexchange.com",
            "linguistics.stackexchange.com",
            "hermeneutics.stackexchange.com",
            "history.stackexchange.com",
            "bricks.stackexchange.com",
            "spanish.stackexchange.com",
            "scicomp.stackexchange.com",
            "movies.stackexchange.com",
            "chinese.stackexchange.com",
            "biology.stackexchange.com",
            "poker.stackexchange.com",
            "mathematica.stackexchange.com",
            "psychology.stackexchange.com",
            "outdoors.stackexchange.com",
            "martialarts.stackexchange.com",
            "sports.stackexchange.com",
            "academia.stackexchange.com",
            "cs.stackexchange.com",
            "workplace.stackexchange.com",
            "chemistry.stackexchange.com",
            "chess.stackexchange.com",
            "raspberrypi.stackexchange.com",
            "russian.stackexchange.com",
            "islam.stackexchange.com",
            "salesforce.stackexchange.com",
            "patents.stackexchange.com",
            "genealogy.stackexchange.com",
            "robotics.stackexchange.com",
            "expressionengine.stackexchange.com",
            "politics.stackexchange.com",
            "anime.stackexchange.com",
            "magento.stackexchange.com",
            "ell.stackexchange.com",
            "sustainability.stackexchange.com",
            "tridion.stackexchange.com",
            "reverseengineering.stackexchange.com",
            "networkengineering.stackexchange.com",
            "opendata.stackexchange.com",
            "freelancing.stackexchange.com",
            "blender.stackexchange.com",
            "mathoverflow.net",
            "space.stackexchange.com",
            "sound.stackexchange.com",
            "astronomy.stackexchange.com",
            "tor.stackexchange.com",
            "pets.stackexchange.com",
            "ham.stackexchange.com",
            "italian.stackexchange.com",
            "pt.stackoverflow.com",
            "aviation.stackexchange.com",
            "ebooks.stackexchange.com",
            "alcohol.stackexchange.com",
            "softwarerecs.stackexchange.com",
            "arduino.stackexchange.com",
            "expatriates.stackexchange.com",
            "matheducators.stackexchange.com",
            "earthscience.stackexchange.com",
            "joomla.stackexchange.com",
            "datascience.stackexchange.com",
            "puzzling.stackexchange.com",
            "craftcms.stackexchange.com",
            "buddhism.stackexchange.com",
            "hinduism.stackexchange.com",
            "communitybuilding.stackexchange.com",
            "worldbuilding.stackexchange.com",
            "ja.stackoverflow.com",
            "emacs.stackexchange.com",
            "hsm.stackexchange.com",
            "economics.stackexchange.com",
            "lifehacks.stackexchange.com",
            "engineering.stackexchange.com",
            "coffee.stackexchange.com",
            "vi.stackexchange.com",
            "musicfans.stackexchange.com",
            "woodworking.stackexchange.com",
            "civicrm.stackexchange.com",
            "medicalsciences.stackexchange.com",
            "ru.stackoverflow.com",
            "rus.stackexchange.com",
            "mythology.stackexchange.com",
            "law.stackexchange.com",
            "opensource.stackexchange.com",
            "elementaryos.stackexchange.com",
            "portuguese.stackexchange.com",
            "computergraphics.stackexchange.com",
            "hardwarerecs.stackexchange.com",
            "es.stackoverflow.com",
            "3dprinting.stackexchange.com",
            "ethereum.stackexchange.com",
            "latin.stackexchange.com",
            "languagelearning.stackexchange.com",
            "retrocomputing.stackexchange.com",
            "crafts.stackexchange.com",
            "korean.stackexchange.com",
            "monero.stackexchange.com",
            "ai.stackexchange.com",
            "esperanto.stackexchange.com",
            "sitecore.stackexchange.com",
            "iot.stackexchange.com",
            "literature.stackexchange.com",
            "vegetarianism.stackexchange.com",
            "ukrainian.stackexchange.com",
            "devops.stackexchange.com",
            "bioinformatics.stackexchange.com",
            "cseducators.stackexchange.com",
            "interpersonal.stackexchange.com",
            "iota.stackexchange.com",
            "stellar.stackexchange.com",
            "conlang.stackexchange.com",
            "quantumcomputing.stackexchange.com",
            "eosio.stackexchange.com",
            "tezos.stackexchange.com",
            "or.stackexchange.com",
            "drones.stackexchange.com",
            "mattermodeling.stackexchange.com",
            "cardano.stackexchange.com",
            "proofassistants.stackexchange.com",
            "substrate.stackexchange.com",
            "bioacoustics.stackexchange.com",
            "solana.stackexchange.com",
            "langdev.stackexchange.com",
            "genai.stackexchange.com",
        ]
        self.domain_guess_mandatory = True

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.user = self.User(self.session)

        self.forum_threads_expr = reliq.expr(
            r'a .s-link href=a>/questions/ | "%(href)v\n"'
        )

        self.guesslist = [
            {
                "func": "get_thread",
                "exprs": [r"/questions/(\d+)"],
            },
            {
                "func": "get_user",
                "exprs": [r"/users/(\d+)"],
            },
            {"func": "get_forum", "exprs": None},
        ]

    def get_next_page(self, rq):
        return rq.search(r'[0] div .pager-answers; [0] a rel="next" href | "%(href)v"')

    def get_forum_next_page(self, rq):
        return rq.search(r'div .pager; [0] a rel=next href | "%(href)v"')

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, rq, ref, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                    .threads div #b>question-summary-; {
                        div .s-post-summary--stats; div .s-post-summary--stats-item; {
                            .score.u [0] * self@ title=b>"Score of ",
                            .views.u [0] * self@ title=e>" views" | "%(title)v",
                            [0] span m@f>"answers"; {
                                .answers.u [0] span .e>-number spre@ | "%i",
                                .solved.b * .has-accepted-answer parent@ | "t"
                            },
                            .bounty.u * self@ .has-bounty | "%i"
                        },
                        div .s-post-summary--content; {
                            * .s-post-summary--content-title; [0] a; {
                                .title * self@ | "%Di" trim,
                                .link * self@ | "%(href)v"
                            },
                            .excerp * .s-post-summary--content-excerpt | "%i",
                            [0] * .s-post-summary--meta; {
                                .tags.a a .s-tag | "%i\n",
                                .author div .s-user-card; {
                                    .avatar img .s-avatar--image | "%(src)v",
                                    div .s-user-card--info; {
                                        .name [0] * c@[0] | "%Di" / trim,
                                        .link a | "%(href)v"
                                    },
                                    .reputation.u li .s-user-card--rep; [0] span | "%(title)v %i" / tr ","
                                },
                                .date span .relativetime | "%(title)v"
                            }
                        }
                    } |
                """
            )
        )

        threads = t["threads"]

        for i in threads:
            i["link"] = url_merge(ref, i["link"])
            i["author"]["link"] = url_merge(ref, i["author"]["link"])
            i["author"]["avatar"] = url_merge(ref, i["author"]["avatar"])

        return {
            "format_version": "stackexchange-forum",
            "url": url,
            "threads": threads,
        }



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import copy
import hashlib
import re


def dict_add(dest, src):
    for i in src.items():
        dest[i[0]] = i[1]


def strtosha256(string):
    if isinstance(string, str):
        string = string.encode()

    return hashlib.sha256(string).hexdigest()


def smarttrim(src):
    # turns " \v i am a   \a \n \n \a test  \a " to "i am a test"

    return " ".join(
        src.translate(str.maketrans("\t\n\r\a\v\f\v", "       ", "")).split()
    )


def settings_copy(
    settings,
):  # function necessary because copy.deepcopy is too stupid to handle TextIOWrapper
    ret = copy.copy(settings)

    for i in settings.keys():
        if isinstance(i, (dict, list, set)):
            ret[i] = copy.deepcopy(ret[i])

    return ret


def get_settings(settings, **kwargs):
    ret = settings_copy(settings)

    for i in settings.keys():
        val = kwargs.get(i)
        if val is not None:
            if isinstance(ret[i], dict) and isinstance(val, dict):
                ret[i].update(val)
            else:
                ret[i] = val
    return ret


def url_valid(url, regex=None, base=False, matchwhole=False):
    if url is None or len(url) == 0:
        return

    pattern = r"^https?://(localhost|([a-zA-Z0-9-]{1,256}\.)+[a-zA-Z]{1,6})(:\d+)?"
    baseg = re.search(pattern, url)
    if not baseg:
        return

    rest = url[baseg.end() :]

    if not regex:
        if base:
            return [baseg[0], rest]
        else:
            return rest

    groups = re.search(regex, rest if not matchwhole else url)
    if not groups:
        return
    if base:
        return [baseg, groups]
    return groups


def url_merge(ref, url):
    if url is None or len(url) == 0:
        return

    url = url.replace("&amp;", "&")
    ref = ref.replace("&amp;", "&")

    if ref[:2] == "//":
        ref = "https:" + ref

    refvalid = url_valid(ref, base=True)
    refbase = None
    refprotocol = "https"
    if refvalid is not None:
        refbase = refvalid[0]
        refprotocol = refbase[: refbase.index(":")]

    if url[:2] == "//":
        return refprotocol + ":" + url
    if url_valid(url) is not None or url[:11] == "data:image/":
        return url

    if ref is None or refvalid is None:
        return

    if ref[-3:] == "/./":
        ref = ref[:-2]

    if (url[:1] != "/" and ref[-1:] != "/") or url[:2] == "./":
        if ref.count("/") > 2:
            ref = re.sub(r"[^/]*$", r"", ref)
        if url[:1] == ".":
            url = url[2:]
    elif ref[-1] != "/" or url[:1] == "/":
        ref = refbase

    if ref[-3:] == "/./":
        ref = ref[:-3]
    if ref[-1] == "/":
        ref = ref[:-1]

    if url[:1] == "/":
        url = url[1:]

    return "{}/{}".format(ref, url)


def url_merge_r(ref, url):
    '''Same as url_merge() but instead of returning None on failure returns ""'''
    r = url_merge(ref, url)
    if r is None:
        return ""
    return r


def conv_short_size(string):
    letter = string[-1:]
    num = 0
    try:
        if letter.isdigit():
            num = float(string)
        else:
            num = float(string[:-1])
            if letter == "M" or letter == "m":
                num *= 1000000
            elif letter == "K" or letter == "k":
                num *= 1000
    except ValueError:
        pass
    return int(num)


def conv_curl_header_to_requests(src):
    r = re.search(r"^\s*([A-Za-z0-9_-]+)\s*:(.*)$", src)
    if r is None:
        return None
    return {r[1]: r[2].strip()}


def conv_curl_cookie_to_requests(src):
    r = re.search(r"^\s*([A-Za-z0-9_-]+)\s*=(.*)$", src)
    if r is None:
        return None
    return {r[1]: r[2].strip()}



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

from ..utils import url_valid


def search_cookies(func, arg, cookies):
    for i in cookies.items():
        if func(i[0], arg):
            return True
    return False


def identify_forum(expr, func, arg, rq, cookies):
    if expr is not None and rq is not None:
        t = rq.search(expr)
        if len(t) != 0:
            return True

    if arg and func:
        return search_cookies(func, arg, cookies)

    return False


def identify_phpbb(url, rq, cookies):
    return identify_forum(
        r"""
        i>body #i>phpbb | "t",
        i>meta i>name=i>copyright i>content=ia>phpbb | "t",
        * Ei>(class|id)=i>copyright m@i>"phpbb" | "t"
    """,
        str.startswith,
        "phpbb_",
        rq,
        cookies,
    )


def identify_xenforo1(url, rq, cookies):
    return identify_forum(r'html #XenForo | "t"', None, None, rq, cookies)


def identify_xenforo2(url, rq, cookies):
    return identify_forum(r'html #XF | "t"', None, None, rq, cookies)


def identify_smf1(url, rq, cookies):
    return identify_forum(
        r'* title="Simple Machines Forum" m@>"Powered by SMF 1." | "t"',
        None,
        None,
        rq,
        cookies,
    )


def identify_smf2(url, rq, cookies):
    return identify_forum(
        r"""
        * title=E>"Simple Machines( Forum)?" | "t",
        script m@B>"[^a-zA-Z0-9]smf_" | "t"
    """,
        None,
        None,
        rq,
        cookies,
    )


def identify_xmb(url, rq, cookies):
    return identify_forum(None, str.__eq__, "xmblva", rq, cookies)


def identify_invision(url, rq, cookies):
    return identify_forum(
        r"""
        * Ei>(class|id)=iE>(el)?copyright m@iE>"(invision|IP\.Board)" | "t"
    """,
        str.startswith,
        "ips4_",
        rq,
        cookies,
    )


def identify_stackexchange(url, rq, cookies):
    return identify_forum(
        r'div .header; h3; [0] a href="https://stackexchange.com/sites" m@f>"more stack exchange communities" | "t"',
        None,
        None,
        rq,
        cookies,
    )


def identify_hackernews(url, rq, cookies):
    base, rest = url_valid(url, base=True)
    if base == "https://news.ycombinator.com":
        return True
    return False


def identify_vbulletin(url, rq, cookies):
    return identify_forum(
        r"""
            html #vbulletin_html | "t",
            style #vbulletin_css | "t",
            div #footer_copyright | "t",
            meta content=b>"vBulletin" | "t",
            script src=a>"/vbulletin_" | "t"
        """,
        None,
        None,
        rq,
        cookies,
    )



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import random
import time
import requests
from threading import Lock
from reliq import reliq

from .exceptions import RequestError, AlreadyVisitedError
from .utils import url_merge_r, smarttrim

useragents = {
    "mozilla": [
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36 CVManaged",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 SberBrowser/16.0.2277.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Chromium GOST) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Atom/26.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.186 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 CCleaner/124.0.0.0",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Chromium GOST) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36 Edg/125.0.0.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.134 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.63 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.67 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 Avast/124.0.0.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.3",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    ],
    "google": [
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36 CVManaged",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 SberBrowser/16.0.2277.0",
        "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Chromium GOST) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Atom/26.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.186 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 CCleaner/124.0.0.0",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Chromium GOST) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36 Edg/125.0.0.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.134 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.63 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.67 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 Avast/124.0.0.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.3",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    ],
    "opera": [
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36 OPR/71.0.3770.171",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 OPR/110.0.0.0",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36 OPR/95.0.0.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36 OPR/91.0.4516.20 (Edition Yx 05)",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36 OPR/94.0.0.0 (Edition Yx 05)",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 OPR/109.0.0.0 (Edition Yx)",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 OPR/110.0.0.0 (Edition Yx 05)",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 OPR/109.0.0.0",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 OPR/109.0.0.0 (Edition Yx 05)",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 OPR/109.0.0.0 (Edition Yx GX 03)",
        "Mozilla/5.0 (Windows NT 10.0; WOW64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36 OPR/89.0.4447.51",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36 OPR/89.0.4447.51",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36 OPR/110.0.0.0",
    ],
}


def to_requests_settings(settings):
    ret = {}
    args = ["timeout", "verify", "proxies", "headers", "cookies", "allow_redirects"]
    for i in args:
        ret[i] = settings[i]

    return ret


class Session(requests.Session):
    def __init__(self, **kwargs):
        super().__init__()

        self.lock = Lock()

        s = to_requests_settings(kwargs)
        self.proxies.update(s["proxies"])
        self.headers.update(s["headers"])
        self.cookies.update(s["cookies"])

        self.get_renew(
            clear=False, user_agent=kwargs["headers"].get("User-Agent") is None
        )

    @staticmethod
    def new_useragent():
        r = random.randint(0, 9)
        if r == 0:
            agents = useragents["opera"]
        elif r >= 1 and r <= 3:
            agents = useragents["mozilla"]
        else:
            agents = useragents["google"]

        return agents[random.randint(0, len(agents) - 1)]

    def get_renew(self, clear=True, user_agent=True):
        if clear:
            self.close()

        if user_agent:
            self.headers.update({"User-Agent": self.new_useragent()})

    def get_req_try(self, url, settings, retry=False):
        if not retry:
            if settings["wait"] != 0:
                time.sleep(settings["wait"])
            if settings["wait_random"] != 0:
                time.sleep(random.randint(0, settings["wait_random"] + 1) / 1000)

        if settings["logger"]:
            if isinstance(settings["logger"], list):
                settings["logger"].append(url)
            else:
                print(url, file=settings["logger"])

        return self.get(url, **to_requests_settings(settings))

    def get_req(self, url, settings, state):
        with self.lock:
            if url in state["visited"]:
                raise AlreadyVisitedError(url)

            state["visited"].add(url)

        tries = settings["retries"]
        retry_wait = settings["retry_wait"]

        instant_end_code = [400, 401, 402, 403, 404, 410, 412, 414, 421, 505]

        i = 0
        while True:
            try:
                resp = self.get_req_try(url, settings, i != 0)
            except (
                requests.ConnectTimeout,
                requests.ConnectionError,
                requests.ReadTimeout,
                requests.exceptions.ChunkedEncodingError,
                RequestError,
            ):
                resp = None

            if resp is None or not (
                resp.status_code >= 200 and resp.status_code <= 299
            ):
                if resp is not None and resp.status_code in instant_end_code:
                    raise RequestError(
                        "failed completely {} {}".format(resp.status_code, url)
                    )
                if i >= tries:
                    raise RequestError(
                        "failed {} {}".format(
                            "connection" if resp is None else resp.status_code, url
                        )
                    )
                i += 1
                if retry_wait != 0:
                    time.sleep(retry_wait)
            else:
                return resp

    @staticmethod
    def base(rq, url):
        ref = url
        u = rq.search(r'[0] head; [0] base href=>[1:] | "%(href)v"')
        if u != "":
            u = url_merge_r(url, u)
            if u != "":
                ref = u
        return ref

    def get_html(self, url, settings, state, trim=False, return_cookies=False):
        resp = self.get_req(url, settings, state)

        r = resp.text
        if trim:
            r = smarttrim(r)

        rq = reliq(r)

        ref = self.base(rq, url)

        if return_cookies:
            return (rq, ref, resp.cookies.get_dict())
        return (rq, ref)

    def get_json(self, url, settings, state):
        resp = self.get_req(url, settings, state)
        return resp.json()
